import argparse
import json
import logging
import os
import torch

from lm_evaluation_harness.lm_eval import tasks, evaluator, utils
from LLMCodebook.models.hf_llama.modeling_llama import LlamaForCausalLM
from LLMCodebook.utils.utils import *
from transformers import LlamaTokenizer
from LLMCodebook.evaluator.ppl import PPLMetric
import copy
logging.getLogger("openai").setLevel(logging.WARNING)

mae = lambda x, y: abs(x - y).mean()

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--base_model', type=str, default="huggyllama/llama-7b", help='base model name')
    # parser.add_argument('--base_model', type=str, default="lmsys/vicuna-7b-v1.3", help='base model name')
    parser.add_argument("--model", required=True)
    parser.add_argument("--model_args", default="")
    parser.add_argument('--output_dir', type=str, default="./lora-alpaca", help='output directory')
    parser.add_argument("--tasks", default=None, choices=utils.MultiChoice(tasks.ALL_TASKS))
    parser.add_argument("--provide_description", action="store_true")
    parser.add_argument("--num_fewshot", type=int, default=0)
    parser.add_argument("--batch_size", type=str, default=None)
    parser.add_argument("--max_batch_size", type=int, default=None,
                        help="Maximal batch size to try with --batch_size auto")
    parser.add_argument("--device", type=str, default=None)
    parser.add_argument("--output_path", default=None)
    parser.add_argument("--limit", type=float, default=None,
                        help="Limit the number of examples per task. "
                             "If <1, limit is a percentage of the total number of examples.")
    parser.add_argument("--data_sampling", type=float, default=None)
    parser.add_argument("--no_cache", action="store_true")
    parser.add_argument("--decontamination_ngrams_path", default=None)
    parser.add_argument("--description_dict_path", default=None)
    parser.add_argument("--check_integrity", action="store_true")
    parser.add_argument("--write_out", action="store_true", default=False)
    parser.add_argument("--output_base_path", type=str, default=None)
    parser.add_argument('--train-new', default=False, type=bool, help='train new centroids and assignments')
    parser.add_argument('--block-size-cin', default=16, type=int, help='quantization block size for cin globally')
    parser.add_argument('--n-centroids-cin', default=1048576, type=int, help='Number of centroids for cin globally')

    return parser.parse_args()


def main():
    args = parse_args()
    torch_version = int(torch.__version__.split('.')[1])
    args.torch_version = torch_version

    assert not args.provide_description  # not implemented

    if args.limit:
        print(
            "WARNING: --limit SHOULD ONLY BE USED FOR TESTING. REAL METRICS SHOULD NOT BE COMPUTED USING LIMIT."
        )

    if args.tasks is None:
        task_names = tasks.ALL_TASKS
    else:
        task_names = utils.pattern_match(args.tasks.split(","), tasks.ALL_TASKS)

    print(f"Selected Tasks: {task_names}")

    description_dict = {}
    if args.description_dict_path:
        with open(args.description_dict_path, "r") as f:
            description_dict = json.load(f)


    tokenizer = LlamaTokenizer.from_pretrained(args.base_model)
    model = LlamaForCausalLM.from_pretrained(
        args.base_model,
        low_cpu_mem_usage=True if args.torch_version >=1.9 else False
    )
    
    if args.device != "cpu":
        model.half()
    model.to(args.device)

    Cmodules = ['q_proj', 'k_proj', 'v_proj', 'o_proj', 'up_proj', 'down_proj', 'gate_proj']
    n_samples = 15
    seq_len = 128
    model.zero_grad()
    
    vector_len = 8
    codebook_len = 2**15
    output_dir = "llama7b"+"_"+str(vector_len)+"_"+str(codebook_len)
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    sub_path = output_dir + "/full_method"
    if not os.path.exists(sub_path):
        os.mkdir(sub_path)

    for name, module in model.named_modules():
        if isinstance(module, nn.Linear) and \
            (any(Cmodule in name for Cmodule in Cmodules)):
            

            if \
            '.0.' not in name and \
            '.1.' not in name and \
            '.2.' not in name and \
            '.30.' not in name and \
            '.31.' not in name : 
                if 'up_proj' in name or 'gate_proj' in name or 'down_proj' in name:
                    codebook(
                        module, 'base_model.model.'+name, 
                        sub_path, None, codebook_len, vector_len, False)
                else:
                    codebook(
                        module, 'base_model.model.'+name, 
                        sub_path, None, codebook_len, vector_len, False)
            else:
                if 'up_proj' in name or 'gate_proj' in name or 'down_proj' in name:
                    codebook(
                        module, 'base_model.model.'+name, 
                        sub_path, None, codebook_len, vector_len, False)
                else:
                    codebook(
                        module, 'base_model.model.'+name, 
                        sub_path, None, codebook_len, vector_len, False)
       
    with torch.no_grad():
        ppl = PPLMetric(model, tokenizer, ['wikitext2', 'ptb'], 128, device="cuda")
        print("PPL after codebook: {}".format(ppl))
    

    tokenizer.pad_token_id = 0
    tokenizer.padding_side = "left"

    results = evaluator.simple_evaluate(
        Cmodel=model,
        model=args.model,
        model_args=args.model_args,
        tasks=task_names,
        num_fewshot=args.num_fewshot,
        batch_size=args.batch_size,
        max_batch_size=args.max_batch_size,
        device=args.device,
        no_cache=args.no_cache,
        limit=args.limit,
        description_dict=description_dict,
        decontamination_ngrams_path=args.decontamination_ngrams_path,
        check_integrity=args.check_integrity,
        write_out=args.write_out,
        output_base_path=args.output_base_path,
    )
    

    dumped = json.dumps(results, indent=2)
    print(dumped)

    if args.output_path:
        os.makedirs(os.path.dirname(args.output_path), exist_ok=True)
        with open(args.output_path, "w") as f:
            f.write(dumped)

    batch_sizes = ",".join(map(str, results["config"]["batch_sizes"]))
    print(
        f"{args.model} ({args.model_args}), limit: {args.limit}, provide_description: {args.provide_description}, "
        f"num_fewshot: {args.num_fewshot}, batch_size: {args.batch_size}{f' ({batch_sizes})' if batch_sizes else ''}"
    )
    print(evaluator.make_table(results))


if __name__ == "__main__":
    main()
